"""
Generate Cert and Pub Key via ACME
"""
from pathlib import Path

import datetime
import base64
import random
import asyncio

from pyasn1.codec.der import encoder
import rfc3161ng


from authsign import crypto, __version__

from authsign.acme_signer import AcmeSigner

from authsign.model import SignedHash
from authsign.utils import (
    CERT_DURATION,
    STAMP_DURATION,
    YEARS,
    parse_date,
    is_time_range_valid,
    open_file,
)

from authsign.log import log_assert, log_message, log_failure, log_success


PASSPHRASE = b"passphrase"

renewing = False


# ============================================================================
class Timestamper:
    def __init__(self, certfile=None, url=None):
        self.cert_pem = None
        with open_file(certfile, "rb") as fh_in:
            self.cert_pem = fh_in.read()

        self._timestamper = rfc3161ng.RemoteTimestamper(
            url, certificate=self.cert_pem, hashname="sha256"
        )

    def __call__(self, text):
        tsr = self._timestamper(data=text.encode("ascii"), return_tsr=True)

        tst = tsr.time_stamp_token

        result = encoder.encode(tsr)

        return base64.b64encode(result), rfc3161ng.get_timestamp(tst)


# ============================================================================
class CertKeyPair:
    """Loads a cert + private key from PEM, extracts public key from cert"""

    def __init__(self):
        self.cert_pem = None
        self.cert = None

        self.public_key = None
        self.public_key_pem = None

        self.private_key = None

    def load(
        self, name, certfile, private_key, passphrase=PASSPHRASE, duration=CERT_DURATION
    ):
        log_message("{0}: Loading Cert: {1}".format(name, str(certfile)))
        with open(certfile, "rb") as fh_in:
            self.set_cert(fh_in.read())

        self.public_key = self.cert.public_key()
        self.public_key_pem = crypto.get_public_key_pem(self.public_key)

        log_message("{0}: Loading Private Key: {1}".format(name, str(private_key)))
        with open(private_key, "rb") as fh_in:
            data = fh_in.read()
            self.private_key = crypto.load_private_key(data, passphrase)

        now = datetime.datetime.utcnow()

        log_assert(self.test_keys("Data Signature Test"), "Validating key pair")

        log_assert(
            is_time_range_valid(self.cert.not_valid_before, now, duration)
            and now <= self.cert.not_valid_after,
            "Validating cert still valid",
        )

        return self

    def init_new(self):
        self.private_key = crypto.create_ecdsa_private_key()

        self.public_key = self.private_key.public_key()
        self.public_key_pem = crypto.get_public_key_pem(self.public_key)

        return self

    def set_cert(self, cert_pem):
        self.cert_pem = cert_pem
        if isinstance(cert_pem, str):
            cert_pem = cert_pem.encode("ascii")
        self.cert = crypto.load_cert(cert_pem)

    def test_keys(self, data):
        """Test key pair sign/verify to ensure its valid"""
        signature = crypto.sign(data, self.private_key)
        return crypto.verify(data, signature, self.public_key)


# ============================================================================
class Signer:
    """Signing cert, private, public key generator"""

    def __init__(
        self,
        domain=None,
        email=None,
        port=None,
        staging=True,
        output=None,
        timestamping=None,
        auth_token=None,
        csca_cert=None,
        csca_private_key=None,
    ):
        self.domain = domain
        self.email = email
        self.port = port
        self.staging = staging

        self.auth_token = auth_token

        log_message("Accepting Auth Token: " + str(self.auth_token))

        self.rootpath = Path(output or "./data")
        self.rootpath.mkdir(exist_ok=True)

        if csca_cert and csca_private_key:
            self.csca_signing = CertKeyPair().load(
                "Cross-Signing",
                csca_cert,
                csca_private_key,
                passphrase=None,
                duration=YEARS,
            )
        else:
            self.csca_signing = None

        self.next_update = 0

        self.domain_signing = None

        self.cs_cert_pem = None

        try:
            self.load_key_pair_and_cert()
        except FileNotFoundError:
            log_message(
                "Signing key or cert not found, creating new signing key + cert"
            )
            self.update_signing_key_and_cert()
        except AssertionError:
            log_message(
                "Signing cert expired or not valid, creating new signing key + cert"
            )
            self.update_signing_key_and_cert()

        if not self.domain_signing:
            raise Exception("Could not load domain signing cert + keys")

        self.timestampers = [Timestamper(**ts_data) for ts_data in timestamping]

    def validate_token(self, auth_header):
        if not self.auth_token:
            return True

        if not auth_header or not auth_header.startswith("bearer "):
            return False

        return auth_header.split(" ")[1] == self.auth_token

    def load_key_pair_and_cert(self):
        """Load key pair and cert"""

        self.domain_signing = CertKeyPair().load(
            "Domain Auth", self.rootpath / "cert.pem", self.rootpath / "private-key.pem"
        )

        self.set_next_update_time(self.domain_signing.cert)

        if self.csca_signing:
            cross_signing = CertKeyPair().load(
                "Cross-Signing Cert",
                self.rootpath / "cs-cert.pem",
                self.rootpath / "private-key.pem",
            )

            self.cs_cert_pem = cross_signing.cert_pem

            log_assert(
                cross_signing.public_key_pem == self.domain_signing.public_key_pem,
                "Cross-Signing Cert Public Key == Domain Cert Public Key",
            )

    def set_next_update_time(self, cert):
        next_update = cert.not_valid_before + CERT_DURATION
        log_message(
            "Certificate will be used from {0} to {1}".format(
                cert.not_valid_before, next_update
            )
        )
        next_update = (next_update - datetime.datetime.utcnow()).total_seconds()
        self.next_update = next_update

    def save_key_pair_and_cert(self):
        """Save keypair and cert"""
        log_message("Saving: " + str(self.rootpath / "private-key.pem"))
        with open(self.rootpath / "private-key.pem", "wb") as fh_out:
            fh_out.write(
                crypto.save_private_key(self.domain_signing.private_key, PASSPHRASE)
            )

        log_message("Saving: " + str(self.rootpath / "cert.pem"))
        with open(self.rootpath / "cert.pem", "wt") as fh_out:
            fh_out.write(self.domain_signing.cert_pem)

        if self.cs_cert_pem:
            log_message("Saving: " + str(self.rootpath / "cs-cert.pem"))
            with open(self.rootpath / "cs-cert.pem", "wt") as fh_out:
                fh_out.write(self.cs_cert_pem)

    def update_signing_key_and_cert(self):
        """Run cert creation"""

        self.domain_signing = CertKeyPair().init_new()

        csr = crypto.create_csr(self.domain, self.domain_signing.private_key)
        csr_pem = crypto.get_as_pem(csr)

        log_message("Awaiting new cert for domain: " + self.domain)

        signer = AcmeSigner(self.domain, self.email, self.port, self.staging)

        try:
            self.domain_signing.set_cert(signer.get_acme_cert(csr_pem))

            log_success("Obtained new domain cert for: " + self.domain)
        except Exception as e:
            log_failure("Unable to retrieve cert for: " + self.domain)
            log_failure("Reason: " + repr(e))
            self.domain_signing = None
            return

        if self.csca_signing:
            now = datetime.datetime.utcnow()

            cs_cert = crypto.create_signed_cert(
                csr,
                self.csca_signing.cert,
                self.csca_signing.private_key,
                now,
                now + CERT_DURATION,
            )
            self.cs_cert_pem = crypto.get_as_pem(cs_cert)

        self.save_key_pair_and_cert()
        self.set_next_update_time(self.domain_signing.cert)

    def __call__(self, sign_req):
        signature = crypto.sign(sign_req.hash, self.domain_signing.private_key)

        timestamper = random.choice(self.timestampers)

        time_signature, timestamp = timestamper(signature)

        created_dt = parse_date(sign_req.created)

        if not is_time_range_valid(created_dt, timestamp, STAMP_DURATION):
            msg = "Created timestamp is out of range: Must be between {0} and {1}, but is {2}".format(
                timestamp, timestamp + STAMP_DURATION, created_dt
            )
            raise Exception(msg)

        return SignedHash(
            software="authsigner " + __version__,
            hash=sign_req.hash,
            created=sign_req.created,
            signature=signature,
            timeSignature=time_signature,
            domain=self.domain,
            domainCert=self.domain_signing.cert_pem,
            timestampCert=timestamper.cert_pem,
            crossSignedCert=self.cs_cert_pem,
        )

    async def renew_loop(self):
        log_message(
            "Signer: Renewing domain certificate in {0}".format(
                datetime.timedelta(seconds=self.next_update)
            )
        )
        loop = asyncio.get_event_loop()
        await asyncio.sleep(self.next_update)
        update_time = CERT_DURATION.total_seconds()
        global renewing

        while True:
            log_message("Signer: Running domain certificate update...")
            renewing = True
            await loop.run_in_executor(None, self.update_signing_key_and_cert)
            log_message(
                "Signer: Renew complete, next renew in {0}".format(
                    datetime.timedelta(seconds=update_time)
                )
            )
            renewing = False
            await asyncio.sleep(update_time)



import logging

logger = logging.getLogger("authsign")

RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
CHECK = "\u2713"
X = "\u2717"


def log_assert(test, msg):
    if test:
        log_success(msg)
    else:
        log_failure(msg)
    assert test


def log_message(text):
    logger.info("  " + (COLOR_SEQ % 33) + text + RESET_SEQ)


def log_success(text):
    logger.info("  " + (COLOR_SEQ % 32) + CHECK + " " + text + RESET_SEQ)


def log_failure(text):
    logger.info("  " + (COLOR_SEQ % 31) + X + " " + text + RESET_SEQ)


def debug_error(text):
    logger.debug("  " + (COLOR_SEQ % 31) + X + " " + text + RESET_SEQ)



from typing import Optional
from pydantic import BaseModel


class SignReq(BaseModel):
    """Sign Request consisting of hash and created date"""

    hash: str
    created: str


class SignedHash(SignReq):
    """Signed Hash of the SignReq, created by signer, ready for verification"""

    version = "0.1.0"

    software: Optional[str]

    signature: str
    domain: str
    domainCert: str
    crossSignedCert: Optional[str]

    timeSignature: str
    timestampCert: str



""" Verify signed responses api"""


import base64
import traceback

import rfc3161ng

from authsign.utils import (
    CERT_DURATION,
    STAMP_DURATION,
    is_time_range_valid,
    parse_date,
    load_yaml,
)
from authsign import crypto
from authsign.log import log_assert, log_message, debug_error
from authsign.model import SignedHash
from authsign.utils import format_date


DEFAULT_TRUSTED_ROOTS = "pkg://authsign.trusted/roots.yaml"


# ============================================================================
class Verifier:
    """Verifies signed response from signer to check for validity"""

    def __init__(self, trusted_roots_filename=None):
        trusted_roots_filename = trusted_roots_filename or DEFAULT_TRUSTED_ROOTS
        log_message("Loading trusted roots from: " + trusted_roots_filename)
        trusted_roots = load_yaml(trusted_roots_filename)

        self.domain_cert_roots = trusted_roots["domain_cert_roots"]
        self.timestamp_cert_roots = trusted_roots["timestamp_cert_roots"]

        log_message(
            "{0} Domain Cert Root(s) Loaded".format(len(self.domain_cert_roots))
        )
        log_message(
            "{0} Timestamp Cert Root(s) Loaded".format(len(self.timestamp_cert_roots))
        )

    def timestamp_verify(self, text, signature, cert_pem):
        """Verify RFC 3161 timestamp given a cert, signature and text
        Return the timestamp"""
        resp = rfc3161ng.decode_timestamp_response(base64.b64decode(signature))
        tst = resp.time_stamp_token

        # verify timestamp was signed by the existing cert
        try:
            rfc3161ng.check_timestamp(
                tst,
                certificate=cert_pem.encode("ascii"),
                data=text.encode("ascii"),
                hashname="sha256",
            )
        except Exception:
            debug_error(traceback.format_exc())
            return None

        return rfc3161ng.get_timestamp(tst)

    def check_fingerprint(self, cert, trusted, name):
        """Check if cert fingerprint matches one of trusted fingerprints (sha-256 hashes)"""
        fingerprint = crypto.get_fingerprint(cert)

        log_assert(
            fingerprint in trusted,
            "Trusted {0} Root Cert (sha-256 fingerprint: {1})".format(
                name, fingerprint
            ),
        )

    def __call__(self, signed_req):
        """Verify signed hash request"""

        if isinstance(signed_req, dict):
            signed_req = SignedHash(**signed_req)

        try:
            log_message("Signing Software: " + str(signed_req.software))

            certs = crypto.validate_cert_chain(signed_req.domainCert.encode("ascii"))
            log_assert(certs, "Verify certificate chain for domain certificate")
            cert = certs[0]

            self.check_fingerprint(certs[-1], self.domain_cert_roots, "Domain")

            public_key = cert.public_key()
            log_assert(
                crypto.verify(signed_req.hash, signed_req.signature, public_key),
                "Verify signature of hash with public key from domain certificate",
            )

            if signed_req.crossSignedCert:
                cs_certs = crypto.validate_cert_chain(
                    signed_req.crossSignedCert.encode("ascii")
                )
                log_assert(
                    cs_certs, "Verify certificate chain for cross-signed certificate"
                )
                cs_public_key = cs_certs[0].public_key()

                log_assert(
                    crypto.verify(signed_req.hash, signed_req.signature, cs_public_key),
                    "Verify signature of hash with public key of cross-signed certificate",
                )

            domain = crypto.get_cert_subject_name(cert)
            log_assert(
                domain == signed_req.domain, "Domain Cert Matches Expected: " + domain
            )

            created = parse_date(signed_req.created)
            log_assert(created, "Parsed signature date")

            log_assert(
                is_time_range_valid(cert.not_valid_before, created, CERT_DURATION),
                "Verify domain certificate was created within '{0}' of creation date".format(
                    str(CERT_DURATION)
                ),
            )

            timestamp = self.timestamp_verify(
                signed_req.signature, signed_req.timeSignature, signed_req.timestampCert
            )

            log_assert(
                timestamp,
                "Verify timeSignature is a valid timestamp signature of\
 hash signature with timestamp certificate",
            )

            log_assert(
                is_time_range_valid(created, timestamp, STAMP_DURATION),
                "Verify time signature created within '{0}' of creation date".format(
                    str(STAMP_DURATION)
                ),
            )

            timestamp_certs = crypto.validate_cert_chain(
                signed_req.timestampCert.encode("ascii")
            )

            log_assert(
                timestamp_certs, "Verify certificate chain for timestamp certificate"
            )

            self.check_fingerprint(
                timestamp_certs[-1], self.timestamp_cert_roots, "Timestamp"
            )

            return {"observer": domain, "timestamp": format_date(timestamp)}

        except Exception:
            debug_error(traceback.format_exc())
            return None



import asyncio
import os

from fastapi import FastAPI, HTTPException, Header

from authsign.signer import Signer
from authsign.verifier import Verifier
from authsign.model import SignedHash, SignReq

from authsign.utils import load_yaml

from authsign.log import log_message, log_failure


loop = asyncio.get_event_loop()
app = FastAPI()

signer = None
verifier = None


@app.on_event("startup")
async def startup_event():
    configfile = os.environ.get("CONFIG", "config.yaml")

    global signer
    log_message("Loading config from: " + configfile)

    config = load_yaml(configfile)

    if os.environ.get("DOMAIN_OVERRIDE"):
        config["signing"]["domain"] = os.environ.get("DOMAIN_OVERRIDE")

    if os.environ.get("PORT_OVERRIDE"):
        config["signing"]["port"] = int(os.environ.get("PORT_OVERRIDE"))

    if os.environ.get("AUTH_TOKEN"):
        config["signing"]["auth_token"] = os.environ.get("AUTH_TOKEN")

    log_message("")
    log_message("Signer init...")
    signer = Signer(**config["signing"])

    if not os.environ.get("NO_RENEW"):
        asyncio.ensure_future(signer.renew_loop())

    global verifier
    log_message("")
    log_message("Verifier Init...")
    verifier = Verifier(config.get("trusted_roots"))
    log_message("")


@app.post("/sign", response_model=SignedHash, response_model_exclude_none=True)
async def sign_data(sign_req: SignReq, authorization: str = Header(None)):
    log_message("Signing Request...")
    if not signer.validate_token(authorization):
        log_failure("Invalid Auth Token")
        raise HTTPException(status_code=403, detail="Invalid auth token")

    try:
        return signer(sign_req)
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.post("/verify")
async def verify_data(signed_hash: SignedHash):
    log_message("Verifying Signed Request...")
    result = await loop.run_in_executor(None, verifier, signed_hash)
    if result:
        return result

    raise HTTPException(status_code=400, detail="Not verified")



import datetime
import importlib
import contextlib

import yaml

# no limit on CA cert validity
YEARS = datetime.timedelta(weeks=1000)

CERT_DURATION = datetime.timedelta(hours=48)

STAMP_DURATION = datetime.timedelta(minutes=10)

ISO_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


def is_time_range_valid(base, thedate, duration):
    return base <= thedate and thedate - base <= duration


def parse_date(datestr):
    try:
        return datetime.datetime.strptime(datestr, ISO_FORMAT)
    except:
        return None


def format_date(date):
    return date.strftime(ISO_FORMAT)


@contextlib.contextmanager
def open_file(filename_or_resource, mode):
    res = None
    if filename_or_resource.startswith("pkg://"):
        pkg, resource = filename_or_resource[6:].split("/", 1)
        if "b" in mode:
            res = importlib.resources.open_binary(pkg, resource)
        else:
            res = importlib.resources.open_text(pkg, resource)
    else:
        res = open(filename_or_resource, mode)

    yield res

    res.close()


def load_yaml(filename):
    with open_file(filename, "rt") as fh:
        data = yaml.load(fh.read(), Loader=yaml.SafeLoader)

    return data



"""
Use ACME protocol to obtain a cert!
"""

from contextlib import contextmanager

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend

import josepy as jose

from acme import challenges
from acme import client
from acme import messages
from acme import standalone


USER_AGENT = "acme-signer"

ACME_STAGING_URL = "https://acme-staging-v02.api.letsencrypt.org/directory"
ACME_PROD_URL = "https://acme-v02.api.letsencrypt.org/directory"


class AcmeSigner:
    """Acme Signer"""

    def __init__(self, csr_pem, email, port, staging=True):
        self.csr_pem = csr_pem
        self.email = email
        self.port = port

        if isinstance(staging, str):
            self.directory_url = staging
        else:
            self.directory_url = ACME_STAGING_URL if staging else ACME_PROD_URL

        self.user_agent = USER_AGENT
        self.staging = bool(staging)

    def create_rsa_key(self):
        """Create RSA Key for ACME auth request"""
        return rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )

    def get_acme_cert(self, csr_pem):
        """Get a signed cert via ACME"""
        # Register account and accept TOS
        acc_key = jose.jwk.JWKRSA(key=self.create_rsa_key())

        net = client.ClientNetwork(
            acc_key, user_agent=self.user_agent, verify_ssl=not self.staging
        )
        directory = messages.Directory.from_json(net.get(self.directory_url).json())
        client_acme = client.ClientV2(directory, net=net)

        # Terms of Service URL is in client_acme.directory.meta.terms_of_service
        # Creates account with contact information.
        client_acme.new_account(
            messages.NewRegistration.from_data(
                email=self.email, terms_of_service_agreed=True
            )
        )

        orderr = client_acme.new_order(csr_pem)

        # Select HTTP-01 within offered challenges by the CA server
        challb = self.select_http01_chall(orderr)

        # The certificate is ready to be used in the variable "fullchain_pem".
        result = self.perform_http01(client_acme, challb, orderr)
        return result.fullchain_pem

    def select_http01_chall(self, orderr):
        """Extract authorization resource from within order resource."""
        # Authorization Resource: authz.
        # This object holds the offered challenges by the server and their status.
        authz_list = orderr.authorizations

        for authz in authz_list:
            # Choosing challenge.
            # authz.body.challenges is a set of ChallengeBody objects.
            for i in authz.body.challenges:
                # Find the supported challenge.
                if isinstance(i.chall, challenges.HTTP01):
                    return i

        raise Exception("HTTP-01 challenge was not offered by the CA server.")

    @contextmanager
    def challenge_server(self, http_01_resources):
        """Manage standalone server set up and shutdown."""

        servers = None
        try:
            servers = standalone.HTTP01DualNetworkedServers(
                ("", self.port), http_01_resources
            )
            # Start client standalone web server.
            servers.serve_forever()
            yield servers
        finally:
            # Shutdown client web server and unbind from PORT
            servers.shutdown_and_server_close()

    def perform_http01(self, client_acme, challb, orderr):
        """Set up standalone webserver and perform HTTP-01 challenge."""

        response, validation = challb.response_and_validation(client_acme.net.key)

        resource = standalone.HTTP01RequestHandler.HTTP01Resource(
            chall=challb.chall, response=response, validation=validation
        )

        with self.challenge_server({resource}):
            # Let the CA server know that we are ready for the challenge.
            client_acme.answer_challenge(challb, response)

            # Wait for challenge status and then issue a certificate.
            # It is possible to set a deadline time.
            finalized_orderr = client_acme.poll_and_finalize(orderr)

        return finalized_orderr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

import pathlib
from setuptools import setup, find_packages

from authsign import __version__


def read(*names, **kwargs):
    with open(pathlib.Path(".").joinpath(*names), "r") as fh:
        return fh.read()


setup(
    name="authsign",
    version=__version__,
    description="Authenticating Data Signing + Verification Server",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Webrecorder Software",
    author_email="info@webrecorder.net",
    url="https://github.com/ikreymer/authsign",
    packages=find_packages(exclude=["tests"]),
    install_requires=[
        line.strip()
        for line in read("requirements.txt").splitlines()
        if not line.strip().startswith("#")
    ],
    test_requires=["pytest", "pytest-asyncio"],
    zip_safe=True,
    package_data={"authsign.trusted": ["*"]},
    data_files={
        "requirements.txt": "requirements.txt",
        "log.json": "log.json",
    },
    entry_points="""
        [console_scripts]
    """,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ],
    python_requires=">=3.7",
)



""" crypto utils"""

import base64
import binascii
import traceback

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec, padding, rsa
from cryptography.hazmat.primitives import hashes

from cryptography.hazmat.backends import default_backend

from cryptography import x509
from cryptography.x509.oid import NameOID

import pem

from authsign.log import debug_error


def create_ecdsa_private_key():
    """Get ECDSA Key"""
    return ec.generate_private_key(ec.SECP256R1(), default_backend())


def create_csr(domain, private_key):
    """Create CSR"""
    builder = (
        x509.CertificateSigningRequestBuilder()
        .subject_name(x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, domain)]))
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName(domain)]), critical=False
        )
    )

    return builder.sign(private_key, hashes.SHA256(), backend=default_backend())


def get_as_pem(csr):
    """Convert a csr or cert object to PEM"""
    return csr.public_bytes(serialization.Encoding.PEM).decode("ascii")


def create_signed_cert(csr, ca_cert, private_ca_key, start_date, end_date):
    """Return a signed certificate from a CSR, using a CA cert + private key"""
    builder = (
        x509.CertificateBuilder()
        .subject_name(csr.subject)
        .issuer_name(ca_cert.issuer)
        .public_key(csr.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(start_date)
        .not_valid_after(end_date)
    )

    return builder.sign(private_ca_key, hashes.SHA256())


def load_cert(cert_pem):
    """Load cert from PEM"""
    return x509.load_pem_x509_certificate(cert_pem, backend=default_backend())


def get_cert_subject_name(cert):
    """Get the subject name (domain) from a cert"""
    return cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value


def get_fingerprint(cert):
    """Get the cert fingerprint as SHA-256 hex string"""
    return binascii.b2a_hex(cert.fingerprint(hashes.SHA256())).decode("ascii")


def get_public_key_pem(public_key):
    """Get PEM for public key"""
    return public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    ).decode("ascii")


def save_private_key(private_key, passphrase):
    """Get PEM of encrypted private key"""
    return private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.BestAvailableEncryption(passphrase),
    )


def load_private_key(pem_data, passphrase):
    """Load private key from PEM"""
    return serialization.load_pem_private_key(
        pem_data, password=passphrase, backend=default_backend()
    )


def load_public_key(pem_data):
    """Load public key from PEM"""
    return serialization.load_pem_public_key(pem_data, backend=default_backend())


def sign(data, private_key):
    """Sign with private_key, return base64-encoded DER"""
    data = private_key.sign(data.encode("ascii"), ec.ECDSA(hashes.SHA256()))
    return base64.b64encode(data).decode("ascii")


def verify(data, signature, public_key):
    """Verify signature (base64-encoded DER) with public key"""
    signature = base64.b64decode(signature)
    data = data.encode("ascii")
    try:
        public_key.verify(signature, data, ec.ECDSA(hashes.SHA256()))
        return True
    except Exception:
        debug_error(traceback.format_exc())
        return False


def validate_cert(cert, public_key):
    """Validation of cert with issuer cert public key (RSA or ECDSA only)
    Does not alone imply the cert is trusted.

    """
    try:
        if isinstance(public_key, rsa.RSAPublicKey):
            public_key.verify(
                cert.signature,
                cert.tbs_certificate_bytes,
                padding.PKCS1v15(),
                cert.signature_hash_algorithm,
            )
        elif isinstance(public_key, ec.EllipticCurvePublicKey):
            public_key.verify(
                cert.signature,
                cert.tbs_certificate_bytes,
                ec.ECDSA(cert.signature_hash_algorithm),
            )

        # only supported RSA and ECDSA certs
        else:
            return False

        return True
    except Exception:
        debug_error(traceback.format_exc())
        return False


def validate_cert_chain(cert_pem):
    """Validate a cert chain stored in PEM file.
    Each cert is validated with key of next cert in PEM file
    Returns all parsed certs, last cert being the root
    """
    prev_cert = None
    certs = []
    for cert in pem.parse(cert_pem):
        cert = load_cert(cert.as_bytes())
        certs.append(cert)
        if prev_cert:
            if not validate_cert(prev_cert, cert.public_key()):
                return None

        prev_cert = cert

    return certs
